const { readFile } = require('fs/promises');
const path = require('path');
const coffeescript = require('coffeescript');

const omit = (obj, keys) =>
  Object.keys(obj)
    .filter((key) => !keys.includes(key))
    .reduce((res, key) => Object.assign(res, { [key]: obj[key] }), {});

const compileCoffee = (code, options) => coffeescript.compile(code, options);

const convertMessage = ({ message, location, code, filename }) => {
  location = {
    file: filename,
    line: location.first_line,
    column: location.first_column,
    length: location.first_line - location.last_column,
    lineText: code,
  };
  return { text: message, location };
};


/**
 * Reads the comments in the file for lines that start with #=
 * Which the proceeding text is used to read the file and collect the text.
 * This happens recursively until no more #= comments are found.
 * 
 * @param {*} path 
 */
const gatherRequireSources = async (accumulated, path, level = 0) => {
  console.log(`gatherRequireSources: ${path} @ ${level}`);
  const sourceText = ""
  const source = await readFile(path, 'utf8');
  const lines = source.split('\n');
  const files = lines
    .filter((line) => {
      if (line.startsWith('#=')) {
        if (!accumulated[path]) {
          return true;
        } else {
          return false
        }
      } else {
        sourceText += line + '\n';
        return false;
      }
    })

    .map((line) => line.replace('#=', ''))
    .map((line) => line.trim())
    // read the file and collect the text
    .map(async (file) => {
      const filePath = path.join(path, file);
      console.log(`gatherRequireSources(${file}) -> ${filePath} @ ${level}`);
      
      const fileSource = await readFile(filePath, 'utf8');
      const result = await gatherRequireSources(accumulated, filePath, level + 1);
      accumulated[file] = true
      sourceText += result;
      return fileSource;
    });A

  return sourceText;

};

const coffeeScriptPlugin = (options = {}) => ({
  name: 'coffeescript',
  setup(build) {
    build.onLoad({ filter: /.\.(coffee|litcoffee)$/ }, async (args) => {
      // const source = await readFile(args.path, 'utf8');
      const accumulated = {}
      const source = await gatherRequireSources(accumulated ,args.path);
      const filename = path.relative(process.cwd(), args.path);
      const opt = omit(options, ['sourceMap']);

      try {
        const contents = compileCoffee(source, { filename, ...opt });
        return {
          contents,
        };
      } catch (e) {
        return {
          errors: [convertMessage(e)],
        };
      }
    });
  },
});

module.exports = coffeeScriptPlugin;
